"""
算法说明：
为了避免扩展散列后重新哈希求key对应的槽，所以采用self.size不变的做法
当负载因子达到设定值（这里设定为0.75）时，将原来的散列保存到self.set中，并新建一个空的散列
此后的元素放在新的散列中。负载因子再次达到设定值时，重复上步骤
在进行查找操作时，即在self.set的每个元素中进行查找
在数据规模不很大时，仍然保持期望的O（1）复杂的
在数据规模非常大时，遍历self.set的时间是线性的，查找总用时也应该是线性的
"""
class HashTable:
    def __init__(self):
        self.size = 11
        # 选取散列槽数为11
        self.slots = [None] * self.size
        self.data = [None] * self.size
        self.factor = 0
        # 引入占用数factor，表示已经被占用的槽数，便于计算负载因子
        self.set = []
        # 这个集合用来记录旧的散列
    def hashfunction(self,key):
        return key % self.size
    def rehash(self,oldhash):
        return (oldhash + 1) % self.size
    def put(self,key,data):
        if self.factor > (0.75 * self.size):
        # 如果负载因子达到设定值（这里是0.75）
        # 就复制旧散列到set中，新建一个槽数仍为11的散列
            self.set.append((self.slots, self.data, self.factor))
            self.slots = [None] * self.size
            self.data = [None] * self.size
            self.factor = 0

        hashvalue = self.hashfunction(key)

        if self.slots[hashvalue] == None:
            self.slots[hashvalue] = key
            self.data[hashvalue] = data
            # 如果这个键不存在于对应槽中,就添加之,并且添加相应的值
        else:
            if self.slots[hashvalue] == key:
                self.data[hashvalue] = data #replace
                # 如果这个键已存在于对应槽中,就用新的值替换原来的值
            else:
                nextslot = self.rehash(hashvalue)
                while self.slots[nextslot] != None and self.slots[nextslot] != key:
                    nextslot = self.rehash(nextslot)
                if self.slots[nextslot] == None:
                    self.slots[nextslot] = key
                    self.data[nextslot] = data
                else:
                    self.data[nextslot] = data # replace
                    # 如果这个键已存在于对应槽中,就用新的值替换原来的值
        self.factor += 1

    def get(self,key):
        startslot = self.hashfunction(key)
        temp = self.set + [(self.slots, self.data, self.factor)]
        # temp包含所有的键值对
        for hashNode in temp:
            data = None
            stop = found = False
            position = startslot
            while hashNode[0][position] != None and not found and not stop:
            # 如果还没找到key或者还没遍历散列，就继续查找
                if hashNode[0][position] == key:
                    found = True
                    # 找到了key，记录之
                    data = hashNode[1][position]
                else:
                    position = self.rehash(position)
                    if position == startslot:
                    # 已经遍历散列而没有找到key，则退出循环
                        stop = True
            if found:
                break
        return data
    def __getitem__(self, key):
        return self.get(key)
    def __setitem__(self, key, data):
        self.put(key,data)
    def __len__(self):
        count = 0
        temp = self.set + [(self.slots, self.data, self.factor)]
        # temp包含所有的键值对
        for x in temp:
            count += x[2]
            # 每次增加对应散列的占用数
        return count
    def __contains__(self, key):
        startslot = self.hashfunction(key)
        temp = self.set + [(self.slots, self.data, self.factor)]
        # temp包含所有的键值对
        for hashNode in temp:
            stop = found = False
            position = startslot
            while hashNode[0][position] != None and not found and not stop:
                # 如果还没找到key或者还没遍历散列，就继续查找
                if hashNode[0][position] == key:
                    found = True
                else:
                    position = self.rehash(position)
                    if position == startslot:
                        # 已经遍历散列而没有找到key，则退出循环
                        stop = True
            if found:
                break
        return found
    def __delitem__(self, key):
        startslot = self.hashfunction(key)
        temp = self.set + [(self.slots, self.data, self.factor)]
        # temp包含所有的键值对
        for hashNode in temp:
            stop = found = False
            position = startslot
            while [position] != None and not found and not stop:
                # 如果还没找到key或者还没遍历散列，就继续查找
                if hashNode[0][position] == key:
                    found = True
                else:
                    position = self.rehash(position)
                    if position == startslot:
                        # 已经遍历散列而没有找到key，则退出循环
                        stop = True
            if found:
                break
        hashNode[0].pop(position)
        # 删除对应的键
        hashNode[1].pop(position)
        # 删除对应的值
# 测试代码
a = HashTable()
a[0] = 'dog'
a[1] = 'cat'
a[11] = 'all'
a[22] = 'like'
a[43] = 'ne'
a[44] = 'ne'
a[45] = 'ne'
a[46] = 'ne'
a[47] = 'ne'
a[48] = 'ne'
a[49] = 'ne'
a[40] = 'ne'
a[50] = 'ne'
a[51] = 'ne'
a[52] = 'ne'
print(52 in a)
del a[52]
print(52 in a)
print(len(a))