class hashNode:
    '''
    散列表的节点
    '''
    def __init__(self, key, data):
        self.data = data
        self.key = key
        self.next = None
    
    def __eq__(self, other):
        return isinstance(other, hashNode) and (other.key==self.key) and (other.data==self.data)
        

class hashMap:
    def __init__(self, num = 11):
        '''
        建立一个空的映射表
        :params num: 散列值的大小，默认为11
        '''
        self.size = num
        # 缓存长度
        self.length = 0
        self.slots = [None] * self.size
        
    def hash_function(self, key):
        '''
        返回一个键的哈希码。使用取余法
        :params key: 要求哈希码的键
        :return: 返回取余的哈希码
        '''
        return key % self.size
    
    def _generate(self, pin = None):
        '''
        生成器母函数
        :params pin: 用一个字符串表示生成的是生成器的某一个属性，None表示本身
        '''
        for slot in self.slots:
            while slot is not None:
                yield slot if pin is None else slot.__getattribute__(pin)
                slot = slot.next
                
    def generate_node(self):
        '''
        返回关于所有节点的生成器
        '''
        return self._generate()        
        
    def generate_key(self):
        '''
        返回所有键的生成器。
        '''
        return self._generate('key')
        
    def generate_value(self):
        '''
        返回所有值的生成器。
        '''
        return self._generate('value')
    
    def generate_hash(self, hash):
        '''
        针对某个hash值进行生成。若hash越界，则抛出IndexError
        '''
        if hash >= self.size or hash < 0:
            raise IndexError('No Such Hash code Here!')
        node = self.slots[hash]
        while node is not None:
            yield node
            node = node.next
                
    def __contains__(self, key):
        '''
        判断键是否在map中
        '''
        return key in self.generate_key()
        
    def __iter__(self):
        '''
        返回迭代器，迭代每一个节点对象
        '''
        return self.generate_node()
        
    def put(self, key, data):
        '''
        放入一个数据
        :params key: 要放入的键
        :params data: 要放入的值
        :return: void
        '''
        hash = self.hash_function(key)
        for node in self.generate_hash(hash):
            if node.key == key:
                node.data = data
                # 长度不加
                return
        # 遍历完毕，发现没有元素，则把元素添加在开头
        node = hashNode(key, data)
        node.next = self.slots[hash]
        self.slots[hash] = node
        # 长度+1
        self.length+=1
        return
        
    def get(self, key):
        '''
        取出一个数据，如果没有这个数据，抛出KeyError
        '''
        hash = self.hash_function(key)
        for node in self.generate_hash(hash):
            if node.key == key:
                return node.data
        raise KeyError(key)
    
    def pop(self, key):
        '''
        取出并移除一个数据。如果没有这个数据，抛出KeyError
        '''
        hash = self.hash_function(key)
        if (self.slots[hash] is not None) and (self.slots[hash].key == key):
            # 判断第一个元素
            node = self.slots[hash]
            self.slots[hash] = node.next
            self.length -= 1
            return node.data
        for node in self.generate_hash(hash):
            if (node.next is not None) and (node.next.key == key):
                # 之所以要下一个，是因为需要连接前一个
                del_node = node.next
                node.next = node.next.next
                self.length -= 1
                return del_node.data
        raise KeyError(key)    
                
    def __len__(self):
        '''
        返回已经占用的长度
        '''
        return self.length
        
    def __str__(self):
        '''
        返回hashmap的字符串表示形式。借用字典的str方法
        '''
        return 'hashMap: %s'%({node.key:node.data for node in self})
        
    def __setitem__(self, key, data):
        '''
        设置散列表的值
        :params key: 键
        :params data: 值
        '''
        self.put(key, data)
        
    def __getitem__(self, key):
        '''
        获取散列表的值。不存在时，抛出KeyError
        :params key: 键
        :return: 返回值
        '''
        try:
            return self.get(key)
        except KeyError as e:
            # 防止生成冗余的异常链
            raise e
            
    def __delitem__(self, key):
        '''
        删除某一个键，不存在时抛出KeyError
        :params key: 键
        '''
        try:
            return self.pop(key)
        except KeyError as e:
            # 防止生成冗余的异常链
            raise e
        
        
if __name__ == '__main__':
    # 测试代码
    m = hashMap()
    print(m)
    for i in range(1,101,11):
        m[i] = i
    print(m)
    print(len(m))
    for i in range(1,101,11):
        del m[i]
    print(m)
    print(len(m))