# -*- coding:utf-8 -*-
class hashMap:
    '''
    本数据结构参考了redis的dict.c中的“分摊转移”思路
    http://www.cppblog.com/IronOxide/archive/2012/08/29/188667.html
    '''
    def __init__(self, size = 11):
        '''
        初始化hashMap，大小为size，在负载因子达到1之后即自动扩充
        :param size: 初始散列表大小
        '''
        self.size = size
        # 原hash表
        self.keys = [None] * size
        self.values = [None] * size
        # 用于扩充时的旧hash表，初始为None
        self.old_keys = None
        self.old_values = None
        # 有效元素总数
        self.length = 0
        
    def put(self, key, value):
        '''
        放入一个元素，使用线性探测解决冲突
        分摊转移，自动解决了扩充问题。
        :param key: 键
        :param value: 值
        :return: 如果新添加了元素则返回True，若修改了某个元素则返回FALSE
        '''
        # 计算hash值
        hash_code = key % self.size
        # 线性探测解决冲突
        while self.keys[hash_code] is not None:
            if key == self.keys[hash_code]:
                # key相同，则直接替换value，并退出
                self.values[hash_code] = value
                return False
            else:
                # 继续寻找下一个
                hash_code = (hash_code + 1) % self.size
        # 冲突消失，放入元素
        self.keys[hash_code] = key
        self.values[hash_code] = value
        self.length += 1
        # 从旧表中转移元素
        if self.old_keys:
            # 旧表不为空，也不为None，可以转移
            old_key = self.old_keys.pop()
            old_value = self.old_values.pop()
            # 计算hash码
            hash_code = old_key % self.size
            while self.keys[hash_code] is not None:
                if key == self.keys[hash_code]:
                    # key相同，由于是旧表，遇到冲突一定不需要替换。
                    # 但是之前对新表添加时多添加了一次，因此需要对长度减一
                    self.length -= 1
                    return False
                else:
                    # 继续寻找下一个
                    hash_code = (hash_code + 1) % self.size
            # 找到位置后，转移过来，也不需要对长度+1
            self.keys[hash_code] = old_key
            self.values[hash_code] = old_value
        # 判断是否满足扩充条件
        if self.size == self.length:
            # 可以证明，此时old_keys和old_values都是满的
            self.old_keys = self.keys
            self.old_values = self.values
            # 新建size为两倍的空表
            self.size *= 2
            self.keys = [None] * self.size
            self.values = [None] * self.size
        # 返回值
        return True
        
    def get(self, key):
        '''
        从散列表中获取元素。若不存在抛出KeyError
        :param key: key
        '''
        # 优先在新散列表中寻找
        if key in self.keys:
            index = self.keys.index(key)
            return self.values[index]
        # 旧散列表中也会寻找
        if key in self.old_keys:
            index = self.old_keys.index(key)
            return self.old_values[index]
        # 都没有，抛出KeyError
        raise KeyError(key)
        
    def pop(self, key):
        '''
        从散列表中移除并返回元素。若不存在抛出KeyError
        :param key: 键
        '''
        # 首先在新散列表中寻找
        if key in self.keys:
            index = self.keys.index(key)
            value = self.values[index]
            # 移除，即把值设置为None
            self.keys[index] = None
            self.values[index] = None
            # 新散列表寻找过后，一定要在旧散列表中进行二次寻找，排除重复元素
            try:
                index = self.old_keys.index(key)
                # 对于旧散列表，维持其满表性质，将元素pop出来
                self.old_keys.pop(index)
                self.old_values.pop(index)
            except ValueError:
                # 没有元素，直接pass
                pass
            self.length -= 1
            return value
        # 旧散列表中也会寻找
        if key in self.old_keys:
            index = self.old_keys.index(key)
            # 对于旧散列表，维持其满表性质，将元素pop出来
            self.old_keys.pop(index)
            self.length -= 1
            return self.old_values.pop(index)
        # 都没有，抛出KeyError
        raise KeyError(key)
        
    def __contains__(self, key):
        '''
        判断一个key是否在map中
        '''
        return key in self.keys or key in self.old_keys
    
    def __len__(self):
        '''
        返回长度
        '''
        return self.length
        
    # 定义魔术方法
    __setitem__ = put
    __getitem__ = get
    __delitem__ = pop
    
    def _generate_all(self):
        '''
        产生一个迭代器，迭代器同时将解决旧表和新表中键的冲突问题
        返回值是元组，(key, value)格式
        '''
        
        for i in range(self.size):
            # 对新表中每个元素
            if self.keys[i] is not None:
                yield (self.keys[i], self.values[i])
        # 对旧表进行处理        
        if self.old_keys:
            for i in range(len(self.old_keys)):
                # 监测新旧表是否有冲突
                if self.old_keys[i] in self.keys:
                    self.old_keys.pop(i)
                    self.old_values.pop(i)
                else:
                    # 没有冲突，则弹出
                    yield (self.old_keys[i], self.old_values[i])
                    
    def __iter__(self):
        '''
        生成迭代器，可以用于遍历元素
        '''
        return self._generate_all()
    
    def __str__(self):
        '''
        返回字符串表示形式。借用dict的str函数
        '''
        return 'hashMap: ' + str(dict(iter(self)))
        
if __name__ == '__main__':
    # 测试代码
    h = hashMap()
    for i in range(25):
        h[i] = i
        print(h)
    for i in reversed(range(25)):
        print(h[i])
        del h[i]
        print(h)
        print(len(h))
